from adminsortable2.admin import SortableAdminMixin
from django.contrib import admin
from django.forms import ModelForm
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from .models import Column, Dataset, Table
from .table import initialize_dataset, setup_fts


@admin.register(Dataset)
class DatasetAdmin(admin.ModelAdmin):
    list_display = ("name", "created_at", "updated_at", "version", "public", "listed")
    list_filter = ("public", "listed")
    search_fields = ("name", "slug")
    readonly_fields = (
        "created_at",
        "updated_at",
    )
    prepopulated_fields = {"slug": ("name",)}
    raw_id_fields = ("default_table",)
    actions = ["initialize_dataset"]

    def save_model(
        self, request: HttpRequest, obj: Dataset, form: ModelForm, change: bool
    ) -> None:
        super().save_model(request, obj, form, change)
        if not obj.tables.exists():
            initialize_dataset(obj)

    @admin.action(description=_("Initialize datasets"))
    def initialize_dataset(self, request: HttpRequest, queryset):
        for dataset in queryset:
            initialize_dataset(dataset)
        self.message_user(request, _("Datasets initialized."))


class ColumnInline(admin.StackedInline):
    model = Column
    extra = 0
    fields = (
        "label",
        "visible",
        "visible_detail",
        "sortable",
        "searchable",
        "facet_count",
        "prefix",
        "postfix",
        "formatter",
        "formatter_arguments",
        "filter",
        "filter_arguments",
    )
    ordering = ("order",)
    show_change_link = True


@admin.register(Table)
class TableAdmin(admin.ModelAdmin):
    inlines = [ColumnInline]
    list_display = ("name", "dataset", "row_count", "visible")
    search_fields = ("name", "slug")
    raw_id_fields = ("primary_key",)
    save_on_top = True
    list_filter = ("dataset", "visible")
    actions = ["generate_fts"]
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "slug",
                    "label",
                    "description",
                    "visible",
                    "row_label_template",
                    "row_description_template",
                    "primary_key",
                    "pagination_size",
                )
            },
        ),
        (
            _("Advanced"),
            {
                "classes": ["collapse"],
                "fields": [
                    "name",
                    "sql",
                    "dataset",
                    "row_count",
                ],
            },
        ),
    )

    @admin.action(description=_("Generate FTS tables"))
    def generate_fts(self, request: HttpRequest, queryset):
        for table in queryset:
            setup_fts(table)


@admin.register(Column)
class ColumnAdmin(SortableAdminMixin, admin.ModelAdmin):
    search_fields = ("name", "label")
    list_filter = ("table", "visible", "visible_detail", "sortable", "searchable")
    list_display = (
        "name",
        "label",
        "table",
        "visible",
        "visible_detail",
        "sortable",
        "searchable",
        "facet_count",
    )

    actions = [
        "make_visible",
        "make_invisible",
        "make_visible_detail",
        "make_invisible_detail",
        "make_sortable",
        "make_unsortable",
        "make_searchable",
        "make_unsearchable",
    ]

    @admin.action(description=_("Make visible"))
    def make_visible(self, request: HttpRequest, queryset):
        queryset.update(visible=True)

    @admin.action(description=_("Make invisible"))
    def make_invisible(self, request: HttpRequest, queryset):
        queryset.update(visible=False)

    @admin.action(description=_("Make visible_detail"))
    def make_visible_detail(self, request: HttpRequest, queryset):
        queryset.update(visible_detail=True)

    @admin.action(description=_("Make invisible_detail"))
    def make_invisible_detail(self, request: HttpRequest, queryset):
        queryset.update(visible_detail=False)

    @admin.action(description=_("Make sortable"))
    def make_sortable(self, request: HttpRequest, queryset):
        queryset.update(sortable=True)

    @admin.action(description=_("Make unsortable"))
    def make_unsortable(self, request: HttpRequest, queryset):
        queryset.update(sortable=False)

    @admin.action(description=_("Make searchable"))
    def make_searchable(self, request: HttpRequest, queryset):
        queryset.update(searchable=True)

    @admin.action(description=_("Make unsearchable"))
    def make_unsearchable(self, request: HttpRequest, queryset):
        queryset.update(searchable=False)



from django import forms
from django.utils.translation import gettext_lazy as _

from .fields import CommaSeparatedMultiChoiceField, RangeField
from .models import FilterChoices

SEARCH_PARAM = "q"
SORT_PARAM = "sort"


class FilterForm(forms.Form):
    def __init__(self, table, *args, **kwargs):
        self.table = table
        super().__init__(*args, **kwargs)
        sortable_columns = [col.name for col in self.table.get_sortable_columns()]
        if sortable_columns:
            self.fields[SORT_PARAM] = CommaSeparatedMultiChoiceField(
                label=_("Sort by"),
                choices=[("", "0")]
                + [(col, col) for col in sortable_columns]
                + [("-" + col, "-" + col) for col in sortable_columns],
                required=False,
                widget=forms.HiddenInput,
            )
        self.set_facet_fields()
        self.set_filter_fields()

    def set_facet_fields(self):
        facet_columns = self.table.get_facet_columns()
        for column in facet_columns:
            self.fields[column.name] = forms.CharField(
                label=column.label,
                required=False,
                widget=forms.HiddenInput(),
            )

    def set_filter_fields(self):
        if self.table.has_fts():
            self.fields[SEARCH_PARAM] = forms.CharField(
                label=_("Search term"),
                required=False,
                widget=forms.TextInput(attrs={"class": "form-control"}),
            )
        filter_columns = self.table.get_filter_columns()
        for column in filter_columns:
            column_filter = column.filter
            if column_filter == FilterChoices.INTEGER_RANGE:
                field_name = column.name + "__range"
                self.fields[field_name] = RangeField(
                    forms.IntegerField,
                    label=column.label,
                    min_value=column.filter_arguments.get("min"),
                    max_value=column.filter_arguments.get("max"),
                    prefix=column.prefix,
                    postfix=column.postfix,
                    required=False,
                    widget=forms.TextInput(
                        attrs={
                            "class": "form-control text-end",
                            "inputmode": "numeric",
                            "type": "number",
                            "min": column.filter_arguments.get("min", ""),
                            "max": column.filter_arguments.get("max", ""),
                        }
                    ),
                )

    def has_filter(self):
        return any(k for k in self.fields.keys() if k != SORT_PARAM)

    def is_filtered(self):
        if not self.is_valid():
            return False
        return any(v for k, v in self.cleaned_data.items() if k != SORT_PARAM)



from django.urls import path
from django.utils.translation import pgettext_lazy

from .views import (
    RowList,
    dataset_index,
    show_dataset_default_table_row,
    show_dataset_table_row,
    table_csv_export,
)

app_name = "datashow"

urlpatterns = [
    path(
        "<slug:slug>/",
        dataset_index,
        name="dataset-index",
    ),
    path(
        pgettext_lazy("url part", "<slug:slug>/table/<slug:table_slug>/"),
        RowList.as_view(),
        name="dataset-table",
    ),
    path(
        pgettext_lazy("url part", "<slug:slug>/table/<slug:table_slug>/export/"),
        table_csv_export,
        name="dataset-table-export",
    ),
    path(
        "<slug:slug>/<slug:row_slug>/",
        show_dataset_default_table_row,
        name="dataset-row",
    ),
    path(
        pgettext_lazy(
            "url part", "<slug:slug>/table/<slug:table_slug>/<slug:row_slug>/"
        ),
        show_dataset_table_row,
        name="dataset-row",
    ),
]



import sqlite3
import tempfile
from contextlib import contextmanager
from pathlib import Path

from django.core.files.base import ContentFile

from .models import Dataset
from .settings import DATASHOW_DB_CACHE_PATH


def get_database_filename(dataset: Dataset) -> str:
    return "datashow-dataset-{}-{}.db".format(dataset.id, dataset.version)


def get_database_file(dataset: Dataset) -> Path:
    filename = get_database_filename(dataset)
    if DATASHOW_DB_CACHE_PATH is None:
        base_path = Path(tempfile.gettempdir())
    else:
        base_path = Path(DATASHOW_DB_CACHE_PATH)

    filepath = base_path / filename
    if filepath.exists():
        return filepath

    base_path.mkdir(exist_ok=True)

    with dataset.sqlite_file.open("rb") as sqlite_file:
        with open(filepath, "wb") as cache_file:
            cache_file.write(sqlite_file.read())

    return filepath


def get_database_connection(dataset: Dataset):
    filepath = get_database_file(dataset)
    return sqlite3.connect("file:{}?mode=ro".format(filepath), uri=True)


@contextmanager
def open_db(dataset):
    connection = get_database_connection(dataset)
    try:
        yield connection
    finally:
        connection.close()


@contextmanager
def open_cursor(dataset):
    with open_db(dataset) as connection:
        cursor = connection.cursor()
        try:
            yield cursor
        finally:
            cursor.close()


@contextmanager
def open_write_cursor(dataset: Dataset):
    filepath = get_database_file(dataset)
    connection = sqlite3.connect(
        "file:{}".format(filepath), uri=True, isolation_level="EXCLUSIVE"
    )
    assert not connection.in_transaction
    try:
        cursor = connection.cursor()

        cursor.execute("PRAGMA JOURNAL_MODE = DELETE;")
        # Smaller page size for better performance when using HTTP range requests
        cursor.execute("PRAGMA page_size = 1024;")
        yield cursor
    finally:
        cursor.execute("COMMIT;")
        cursor.execute("VACUUM;")
        connection.commit()
        connection.close()

    # Write database back to origin and update version
    with open(filepath, "rb") as cache_file:
        new_sqlite = ContentFile(cache_file.read())
    dataset.version += 1
    filename = get_database_filename(dataset)
    dataset.sqlite_file.save(filename, new_sqlite, save=True)



import logging
import sqlite3
from typing import Generator, NamedTuple, Optional

from django.utils.text import slugify

from .db import open_cursor, open_write_cursor
from .formatters import format_value
from .models import Column, Dataset, Table
from .query import NONE_VALUE, SqlQuery

logger = logging.getLogger(__name__)


class SqliteColumn(NamedTuple):
    name: str
    type: str
    notnull: int
    default: str
    pk: int


class SqliteTableData(NamedTuple):
    name: str
    columns: list[SqliteColumn]
    count: int


SQLITE_TYPE_FORMATTER_MAP = {
    "NULL": "plaintext",
    "INTEGER": "integer",
    "REAL": "float",
    "TEXT": "plaintext",
    "BLOB": "plaintext",
}


class FacetEntry(NamedTuple):
    key: str
    count: int
    active: bool


class Facet(NamedTuple):
    column: Column
    keys: list[FacetEntry]


def get_facets(table: Table, formdata: Optional[dict]) -> list[Facet]:
    facet_cols = [col for col in table.get_columns() if col.facet_count]
    facets = []
    for column in facet_cols:
        query = SqlQuery(table, formdata).with_facet(column)
        active_value = None
        if formdata and column.name in formdata:
            active_value = str(formdata[column.name])
        counts = list(get_facet_counts(table, query, active_value))
        facets.append(Facet(column, counts))
    return facets


class RowQueryset:
    def __init__(self, table: Table, formdata):
        self.table = table
        self.formdata = formdata
        self._rows = []

    def count(self):
        query = SqlQuery(self.table, self.formdata).with_count()
        if not query.params:
            if not self.table.row_count:
                self.table.row_count = get_count(self.table, query)
                if self.table.row_count:
                    self.table.save(update_fields=["row_count"])
            return self.table.row_count
        return get_count(self.table, query)

    def stream_raw(self, chunk_size=1000):
        query = SqlQuery(self.table, self.formdata).with_all()
        return get_raw_rows_iter(self.table, query, chunk_size)

    def __getitem__(self, page_slice):
        # Evaluate on slice
        query = SqlQuery(self.table, self.formdata).with_list(page_slice=page_slice)
        self._rows = get_rows(self.table, query)
        return self

    def __iter__(self):
        return iter(self._rows)


def get_table_info(cursor, table_name: str) -> SqliteTableData:
    cursor.execute("PRAGMA table_info({})".format(table_name))
    columns = [SqliteColumn(*t[1:]) for t in cursor.fetchall()]

    cursor.execute("SELECT COUNT(*) FROM {}".format(table_name))
    row_count = cursor.fetchone()[0]

    return SqliteTableData(name=table_name, columns=columns, count=row_count)


def get_sqlite_table_data(dataset: Dataset) -> list[SqliteTableData]:
    with open_cursor(dataset) as cursor:
        cursor.execute("SELECT * FROM sqlite_master WHERE type='table';")
        sqlite_tables = cursor.fetchall()
        filtered_sqlite_tables = [
            table[1]
            for table in sqlite_tables
            if not table[1].endswith("__fts") and "__fts_" not in table[1]
        ]
        return [get_table_info(cursor, table) for table in filtered_sqlite_tables]


def initialize_dataset(dataset: Dataset):
    for sqlite_table_data in get_sqlite_table_data(dataset):
        table, _created = Table.objects.get_or_create(
            dataset=dataset,
            name=sqlite_table_data.name,
            defaults={
                "slug": slugify(sqlite_table_data.name),
                "label": sqlite_table_data.name,
                "row_count": sqlite_table_data.count,
            },
        )
        existing_columns = set(
            Column.objects.filter(table=table).values_list("name", flat=True)
        )
        incoming_columns = set(col.name for col in sqlite_table_data.columns)
        obsolete_columns = existing_columns - incoming_columns
        Column.objects.filter(table=table, name__in=obsolete_columns).delete()

        pk_col = None
        for i, sqlite_column in enumerate(sqlite_table_data.columns):
            column, _created = Column.objects.get_or_create(
                table=table,
                name=sqlite_column.name,
                defaults={
                    "label": sqlite_column.name.capitalize(),
                    "formatter": SQLITE_TYPE_FORMATTER_MAP.get(
                        sqlite_column.type, "plaintext"
                    ),
                    "order": i,
                },
            )
            if sqlite_column.pk and not pk_col:
                pk_col = column
        if pk_col:
            table.primary_key = column
            table.save(update_fields=["primary_key"])


class RenderRow(NamedTuple):
    value: str
    css: str
    formatted_value: str
    column: Column


def format_column_value(row_data, column: Column, detail=False) -> RenderRow:
    value = row_data[column.name]
    css, formatted_value = format_value(column, value, row_data, detail=detail)
    return RenderRow(value, css, formatted_value, column)


def get_count(table: Table, query: SqlQuery):
    try:
        with open_cursor(table.dataset) as cursor:
            cursor.execute(query.to_sql(), query.params)
            return cursor.fetchone()[0]
    except sqlite3.OperationalError as e:
        logger.error("Error getting count: %s with %s", e, query)
        return 0


def get_facet_counts(
    table: Table, query: SqlQuery, active_value: Optional[str] = None
) -> Generator[FacetEntry, None, None]:
    try:
        with open_cursor(table.dataset) as cursor:
            cursor.execute(query.to_sql(), query.params)
            for row in cursor.fetchall():
                key = row[0]
                count = row[1]
                if key is None:
                    key = NONE_VALUE
                else:
                    key = str(key)
                yield FacetEntry(key, count, key == active_value)
    except sqlite3.OperationalError as e:
        logger.error("Error getting facet count: %s with %s", e, query)
        return 0


def get_rows(table: Table, query: SqlQuery) -> list[list[RenderRow]]:
    dataset = table.dataset
    col_list = table.get_sql_columns()
    try:
        with open_cursor(dataset) as cursor:
            cursor.execute(query.to_sql(), query.params)
            rows = cursor.fetchall()
            # Fetch all data
            row_list = [
                dict([(col, val) for col, val in zip(col_list, row)]) for row in rows
            ]
            # Format with complete row data
            # but return only visible columns
            return [
                [
                    format_column_value(row_data, col)
                    for col in table.get_visible_columns()
                ]
                for row_data in row_list
            ]
    except sqlite3.OperationalError as e:
        logger.error("Error getting rows: %s with %s", e, query)
        return []


def get_raw_rows_iter(
    table: Table, query: SqlQuery, chunk_size=1000
) -> Generator[dict, None, None]:
    dataset = table.dataset
    try:
        with open_cursor(dataset) as cursor:
            cursor.execute(query.to_sql(), query.params)
            while True:
                rows = cursor.fetchmany(chunk_size)
                if not rows:
                    break
                for row in rows:
                    yield [str(val) if val is not None else "" for val in row]
    except sqlite3.OperationalError as e:
        logger.error("Error getting rows: %s", e)
        return []


def get_row(table: Table, row_id: str) -> tuple[list[RenderRow], dict]:
    col_list = table.get_sql_columns()
    with open_cursor(table.dataset) as cursor:
        query = SqlQuery(table).with_row(row_id)
        cursor.execute(query.to_sql(), query.params)
        row = cursor.fetchone()
        if not row:
            raise KeyError("Row not found")
        row_data = dict([(col, val) for col, val in zip(col_list, row)])
        # Return only visible columns
        return (
            [
                format_column_value(row_data, col, detail=True)
                for col in table.get_detail_visible_columns()
            ],
            row_data,
        )


def setup_fts(table: Table):
    with open_write_cursor(table.dataset) as cursor:
        setup_fts_with_cursor(cursor, table)


def setup_fts_with_cursor(cursor, table: Table) -> None:
    if not table.has_fts():
        return
    fts_columns = [col.name for col in table.get_columns() if col.searchable]
    fts_table_name = f"{table.name}__fts"
    pk_col = table.primary_key.name
    all_fts_columns = [pk_col] + fts_columns

    cursor.execute(
        "DROP TABLE IF EXISTS {}".format(fts_table_name),
    )
    cursor.execute(
        "CREATE VIRTUAL TABLE {} USING FTS5({})".format(
            fts_table_name, ", ".join(all_fts_columns)
        )
    )
    cursor.execute(
        "INSERT INTO {}({}) SELECT {} FROM {}".format(
            fts_table_name,
            ", ".join(all_fts_columns),
            ", ".join(all_fts_columns),
            table.as_sql(),
        )
    )



# Generated by Django 4.2.16 on 2025-02-24 10:10

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("datashow", "0002_column_facet_count"),
    ]

    operations = [
        migrations.AddField(
            model_name="table",
            name="row_description_template",
            field=models.TextField(
                blank=True,
                help_text="Python format template for describing rows / rendered as Markdown/HTML!",
                verbose_name="Row description template",
            ),
        ),
        migrations.AlterField(
            model_name="column",
            name="formatter",
            field=models.CharField(
                choices=[
                    ("plaintext", "Plaintext"),
                    ("float", "Float"),
                    ("integer", "Integer"),
                    ("date", "Date"),
                    ("datetime", "Date and time"),
                    ("boolean", "Boolean"),
                    ("link", "Link"),
                    ("summary", "Summary"),
                    ("abbreviation", "Abbreviation"),
                    ("iframe", "IFrame"),
                ],
                default="plaintext",
                help_text="Formatter for the column.",
                max_length=255,
                verbose_name="Formatter",
            ),
        ),
    ]



import markdown
from django import template
from django.utils.encoding import force_str
from django.utils.safestring import mark_safe

from ..forms import SORT_PARAM

register = template.Library()


@register.simple_tag(takes_context=True)
def querystring(context, key=None, value=None):
    request = context["request"]
    if key is None:
        return request.META["QUERY_STRING"]
    if value is None:
        query = request.GET.copy()
        query.pop(key, None)
        return query.urlencode()
    if request.GET.get(key) == value:
        return ""

    query = request.GET.copy()
    query[key] = value
    return query.urlencode()


@register.simple_tag(takes_context=True)
def sort_queryparam(context, column, ascending=True, remove=False):
    request = context["request"]
    current_value = request.GET.get(SORT_PARAM, "")
    value = column if ascending else "-" + column

    values_set = dict.fromkeys(x for x in current_value.split(",") if x)
    if not remove and value in values_set:
        return ""

    if remove:
        value = ",".join(x for x in values_set if x != value)
    else:
        values = [x for x in values_set if x != column and x[1:] != column]
        values.append(value)
        value = ",".join(values)

    query = request.GET.copy()
    if not value:
        query.pop(SORT_PARAM, None)
    else:
        query[SORT_PARAM] = value
    return query.urlencode()


@register.filter(name="listify")
def listify(value):
    return list(value)


@register.filter(name="markdownify")
def markdownify(value):
    return mark_safe(markdown.markdown(force_str(value)))



# Generated by Django 4.2.16 on 2025-01-22 08:48

import django.db.models.deletion
from django.db import migrations, models

import datashow.models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Column",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Internal name of the column.",
                        max_length=255,
                        verbose_name="Name",
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        help_text="Displayed label of the column.",
                        max_length=255,
                        verbose_name="Label",
                    ),
                ),
                (
                    "order",
                    models.PositiveIntegerField(
                        default=0,
                        help_text="Order of the column.",
                        verbose_name="Order",
                    ),
                ),
                (
                    "visible",
                    models.BooleanField(
                        default=True,
                        help_text="Indicates whether the column is visible in table view.",
                        verbose_name="Visible",
                    ),
                ),
                (
                    "visible_detail",
                    models.BooleanField(
                        default=True,
                        help_text="Indicates whether the column is visible in detail view.",
                        verbose_name="Visible in detail",
                    ),
                ),
                (
                    "sortable",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether the column is sortable.",
                        verbose_name="Sortable",
                    ),
                ),
                (
                    "searchable",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether the column is included in search queries.",
                        verbose_name="Searchable",
                    ),
                ),
                (
                    "prefix",
                    datashow.models.NonStrippingCharField(
                        blank=True,
                        help_text="Prefix when outputting the column value.",
                        max_length=255,
                        verbose_name="Prefix",
                    ),
                ),
                (
                    "postfix",
                    datashow.models.NonStrippingCharField(
                        blank=True,
                        help_text="Postfix when outputting the column value.",
                        max_length=255,
                        verbose_name="Postfix",
                    ),
                ),
                (
                    "formatter",
                    models.CharField(
                        choices=[
                            ("plaintext", "Plaintext"),
                            ("float", "Float"),
                            ("integer", "Integer"),
                            ("date", "Date"),
                            ("datetime", "Date and time"),
                            ("boolean", "Boolean"),
                            ("link", "Link"),
                            ("summary", "Summary"),
                            ("abbreviation", "Abbreviation"),
                        ],
                        default="plaintext",
                        help_text="Formatter for the column.",
                        max_length=255,
                        verbose_name="Formatter",
                    ),
                ),
                (
                    "formatter_arguments",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Arguments for the formatter.",
                        verbose_name="Formatter Arguments",
                    ),
                ),
                (
                    "filter",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("ilike", "Substring"),
                            ("integer_range", "Integer Range"),
                        ],
                        default="",
                        help_text="Filter function for the column.",
                        max_length=255,
                        verbose_name="Filter",
                    ),
                ),
                (
                    "filter_arguments",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Arguments for the selected filter.",
                        verbose_name="Filter Arguments",
                    ),
                ),
            ],
            options={
                "verbose_name": "Column",
                "verbose_name_plural": "Columns",
                "ordering": ("order",),
            },
        ),
        migrations.CreateModel(
            name="Dataset",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="The name of the dataset.",
                        max_length=255,
                        verbose_name="Name",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="Unique slug for the dataset.",
                        max_length=255,
                        unique=True,
                        verbose_name="Slug",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        help_text="Description for the dataset, default rendered as Markdown/HTML!",
                        verbose_name="Description",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The date and time when the dataset was created.",
                        verbose_name="Created At",
                    ),
                ),
                (
                    "updated_at",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The date and time when the dataset was last updated.",
                        verbose_name="Updated At",
                    ),
                ),
                (
                    "public",
                    models.BooleanField(
                        default=True,
                        help_text="Indicates whether the dataset is publicly accessible.",
                        verbose_name="Is public",
                    ),
                ),
                (
                    "listed",
                    models.BooleanField(
                        default=True,
                        help_text="Indicates whether the dataset is listed.",
                        verbose_name="Is listed",
                    ),
                ),
                (
                    "sqlite_file",
                    models.FileField(
                        help_text="Upload the SQLite file for the dataset.",
                        storage=datashow.models.select_storage,
                        upload_to=datashow.models.get_sqlite_file_storage_path,
                        verbose_name="SQLite File",
                    ),
                ),
                (
                    "version",
                    models.IntegerField(
                        default=1,
                        help_text="The version number of the dataset.",
                        verbose_name="Version",
                    ),
                ),
            ],
            options={
                "verbose_name": "Dataset",
                "verbose_name_plural": "Datasets",
                "ordering": ("updated_at",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Table",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Internal name of the table.",
                        max_length=255,
                        verbose_name="Name",
                    ),
                ),
                ("slug", models.SlugField(max_length=255, verbose_name="Slug")),
                (
                    "label",
                    models.CharField(
                        blank=True,
                        help_text="Displayed label for the table.",
                        max_length=255,
                        verbose_name="Label",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        help_text="Description for the dataset, default rendered as Markdown/HTML!",
                        verbose_name="Description",
                    ),
                ),
                (
                    "row_count",
                    models.IntegerField(
                        default=0,
                        help_text="The number of rows in the table.",
                        verbose_name="Row Count",
                    ),
                ),
                (
                    "visible",
                    models.BooleanField(
                        default=True,
                        help_text="Indicates whether the table is visible.",
                        verbose_name="Visible",
                    ),
                ),
                (
                    "row_label_template",
                    models.CharField(
                        blank=True,
                        help_text="Python format template for labeling rows.",
                        max_length=512,
                        verbose_name="Row label template",
                    ),
                ),
                (
                    "pagination_size",
                    models.PositiveIntegerField(
                        default=10,
                        help_text="The number of rows per page.",
                        verbose_name="Pagination Size",
                    ),
                ),
                (
                    "sql",
                    models.TextField(
                        blank=True,
                        help_text="Makes this table 'virtual' by representing this table with this SQL query.",
                        verbose_name="SQL representation",
                    ),
                ),
                (
                    "dataset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tables",
                        to="datashow.dataset",
                        verbose_name="Dataset",
                    ),
                ),
                (
                    "primary_key",
                    models.ForeignKey(
                        blank=True,
                        help_text="For use in identifying rows in URLs.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="datashow.column",
                        verbose_name="Primary Key Column",
                    ),
                ),
            ],
            options={
                "verbose_name": "Table",
                "verbose_name_plural": "Tables",
                "ordering": ("name",),
            },
        ),
        migrations.AddField(
            model_name="dataset",
            name="default_table",
            field=models.ForeignKey(
                blank=True,
                help_text="The default table is displayed instead of a list of all tables.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="datashow.table",
                verbose_name="Default Table",
            ),
        ),
        migrations.AddField(
            model_name="column",
            name="table",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="columns",
                to="datashow.table",
                verbose_name="Table",
            ),
        ),
        migrations.AddConstraint(
            model_name="table",
            constraint=models.UniqueConstraint(
                fields=("dataset", "slug"), name="unique_table_slug"
            ),
        ),
        migrations.AddConstraint(
            model_name="table",
            constraint=models.UniqueConstraint(
                fields=("dataset", "name"), name="unique_table_name"
            ),
        ),
        migrations.AddConstraint(
            model_name="column",
            constraint=models.UniqueConstraint(
                fields=("table", "name"), name="unique_table_column_name"
            ),
        ),
    ]



from django import forms

# from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _


class RangeWidget(forms.MultiWidget):
    template_name = "datashow/widgets/range.html"

    def __init__(self, widget, *args, **kwargs):
        self.prefix = kwargs.pop("prefix", "")
        self.postfix = kwargs.pop("postfix", "")
        super().__init__(widgets=(widget, widget), *args, **kwargs)
        self.widgets_names = ["_min", "_max"]

    def decompress(self, value):
        return value

    def get_context(self, name, value, attrs):
        ctx = super().get_context(name, value, attrs)
        ctx["prefix"] = self.prefix
        ctx["postfix"] = self.postfix
        return ctx


class RangeField(forms.MultiValueField):
    default_error_messages = {
        "invalid_start": _("Enter a valid start value."),
        "invalid_end": _("Enter a valid end value."),
    }

    def __init__(
        self,
        field_class,
        min_value=None,
        max_value=None,
        prefix="",
        postfix="",
        widget=forms.TextInput,
        *args,
        **kwargs,
    ):
        if "initial" not in kwargs:
            kwargs["initial"] = ["", ""]

        fields = (
            field_class(
                min_value=min_value,
                max_value=max_value,
            ),
            field_class(
                min_value=min_value,
                max_value=max_value,
            ),
        )

        super().__init__(
            fields=fields,
            widget=RangeWidget(widget, prefix=prefix, postfix=postfix),
            *args,
            **kwargs,
        )

    def compress(self, data_list):
        if data_list:
            return [
                self.fields[0].clean(data_list[0]),
                self.fields[1].clean(data_list[1]),
            ]
        return None


class CommaSeparatedMultiChoiceField(forms.MultipleChoiceField):
    default_error_messages = {
        "invalid_start": _("Enter a valid start value."),
        "invalid_end": _("Enter a valid end value."),
    }

    def to_python(self, value):
        if value is None:
            return []
        value = value.split(",")
        return [str(val) for val in value]



# Generated by Django 4.2.16 on 2025-01-23 09:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('datashow', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='column',
            name='facet_count',
            field=models.PositiveSmallIntegerField(default=0, help_text='Number of facet entries, 0 if disabled.', verbose_name='Facet Count'),
        ),
    ]



from django.core.exceptions import ValidationError
from django.core.files.storage import storages
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from .settings import DATASHOW_STORAGE_BACKEND


def get_sqlite_file_storage_path(instance, filename):
    filename = instance.slug + ".db"
    return "datashow/sqlitefiles/{}".format(filename)


def select_storage():
    return storages[DATASHOW_STORAGE_BACKEND]


class AbstractDataset(models.Model):
    name = models.CharField(
        max_length=255,
        verbose_name=_("Name"),
        help_text=_("The name of the dataset."),
    )
    slug = models.SlugField(
        max_length=255,
        unique=True,
        verbose_name=_("Slug"),
        help_text=_("Unique slug for the dataset."),
    )
    description = models.TextField(
        blank=True,
        verbose_name=_("Description"),
        help_text=_("Description for the dataset, default rendered as Markdown/HTML!"),
    )
    created_at = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("Created At"),
        help_text=_("The date and time when the dataset was created."),
    )
    updated_at = models.DateTimeField(
        auto_now=True,
        verbose_name=_("Updated At"),
        help_text=_("The date and time when the dataset was last updated."),
    )
    public = models.BooleanField(
        default=True,
        verbose_name=_("Is public"),
        help_text=_("Indicates whether the dataset is publicly accessible."),
    )
    listed = models.BooleanField(
        default=True,
        verbose_name=_("Is listed"),
        help_text=_("Indicates whether the dataset is listed."),
    )
    sqlite_file = models.FileField(
        upload_to=get_sqlite_file_storage_path,
        storage=select_storage,
        verbose_name=_("SQLite File"),
        help_text=_("Upload the SQLite file for the dataset."),
    )
    version = models.IntegerField(
        default=1,
        verbose_name=_("Version"),
        help_text=_("The version number of the dataset."),
    )
    default_table = models.ForeignKey(
        "Table",
        blank=True,
        on_delete=models.SET_NULL,
        null=True,
        related_name="+",
        verbose_name=_("Default Table"),
        help_text=_("The default table is displayed instead of a list of all tables."),
    )

    class Meta:
        abstract = True
        verbose_name = _("Dataset")
        verbose_name_plural = _("Datasets")
        ordering = ("updated_at",)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("datashow:dataset-index", kwargs={"slug": self.slug})

    def clean(self):
        if self.default_table and self.default_table.dataset_id != self.pk:
            raise ValidationError(
                {"default_table": "Default table needs to be part of this dataset."}
            )


class Dataset(AbstractDataset):
    class Meta(AbstractDataset.Meta):
        abstract = False


class Table(models.Model):
    dataset = models.ForeignKey(
        Dataset,
        on_delete=models.CASCADE,
        related_name="tables",
        verbose_name=_("Dataset"),
    )
    name = models.CharField(
        max_length=255,
        verbose_name=_("Name"),
        help_text=_("Internal name of the table."),
    )
    slug = models.SlugField(
        max_length=255,
        verbose_name=_("Slug"),
    )
    label = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_("Label"),
        help_text=_("Displayed label for the table."),
    )
    description = models.TextField(
        blank=True,
        verbose_name=_("Description"),
        help_text=_("Description for the dataset, default rendered as Markdown/HTML!"),
    )
    row_count = models.IntegerField(
        default=0,
        verbose_name=_("Row Count"),
        help_text=_("The number of rows in the table."),
    )
    visible = models.BooleanField(
        default=True,
        verbose_name=_("Visible"),
        help_text=_("Indicates whether the table is visible."),
    )
    row_label_template = models.CharField(
        _("Row label template"),
        max_length=512,
        blank=True,
        help_text=_("Python format template for labeling rows."),
    )
    row_description_template = models.TextField(
        _("Row description template"),
        blank=True,
        help_text=_(
            "Python format template for describing rows / rendered as Markdown/HTML!"
        ),
    )
    primary_key = models.ForeignKey(
        "Column",
        verbose_name=_("Primary Key Column"),
        help_text=_("For use in identifying rows in URLs."),
        blank=True,
        on_delete=models.SET_NULL,
        null=True,
        related_name="+",
    )
    pagination_size = models.PositiveIntegerField(
        default=10,
        verbose_name=_("Pagination Size"),
        help_text=_("The number of rows per page."),
    )
    sql = models.TextField(
        _("SQL representation"),
        blank=True,
        help_text=_(
            "Makes this table 'virtual' by representing this table with this SQL query."
        ),
    )

    class Meta:
        verbose_name = _("Table")
        verbose_name_plural = _("Tables")
        ordering = ("name",)
        constraints = [
            models.UniqueConstraint(
                fields=["dataset", "slug"], name="unique_table_slug"
            ),
            models.UniqueConstraint(
                fields=["dataset", "name"], name="unique_table_name"
            ),
        ]

    def __str__(self):
        return f"{self.name} ({self.dataset.name})"

    def clean(self):
        if self.primary_key and self.primary_key.table_id != self.pk:
            raise ValidationError(
                _("Primary Key Column needs to be part of this table.")
            )
        example_row = {k: k for k in self.get_sql_columns()}
        try:
            self.row_label(example_row)
        except (KeyError, ValueError) as e:
            raise ValidationError(
                _("Row label template formatting error: %s") % e
            ) from e
        try:
            self.row_description(example_row)
        except (KeyError, ValueError) as e:
            raise ValidationError(
                _("Row description template formatting error: %s") % e
            ) from e

    def get_absolute_url(self):
        if self.dataset.default_table_id == self.pk:
            return self.dataset.get_absolute_url()
        return reverse(
            "datashow:dataset-table",
            kwargs={"slug": self.dataset.slug, "table_slug": self.slug},
        )

    def as_sql(self):
        """
        Table can be defined by a SQL query or by an actual table with that name.
        """
        if self.sql:
            return '({}) AS "{}"'.format(self.sql, self.name)
        return self.name

    def get_row_count(self):
        from .query import SqlQuery
        from .table import get_count

        query = SqlQuery(self, None).with_count()
        return get_count(self, query)

    def get_columns(self) -> list["Column"]:
        if not hasattr(self, "_columns"):
            self._columns = list(self.columns.all())
        return self._columns

    def get_visible_columns(self) -> list["Column"]:
        return [col for col in self.get_columns() if col.visible]

    def get_detail_visible_columns(self) -> list["Column"]:
        return [col for col in self.get_columns() if col.visible_detail]

    def get_filter_columns(self) -> list["Column"]:
        return [col for col in self.get_columns() if col.filter]

    def get_facet_columns(self) -> list["Column"]:
        return [col for col in self.get_columns() if col.facet_count]

    def get_sortable_columns(self) -> list["Column"]:
        return [col for col in self.get_columns() if col.sortable]

    def get_sql_columns(self) -> list[str]:
        return [col.name for col in self.get_columns()]

    def get_formatted_columns(self) -> list[dict]:
        from .formatters import format_column

        return [(c, format_column(c)) for c in self.get_visible_columns()]

    def has_fts(self):
        if not self.primary_key:
            return False
        return any(col.searchable for col in self.get_columns())

    def row_label(self, row: dict) -> str:
        try:
            return self.row_label_template.format(**row)
        except (KeyError, ValueError):
            return ""

    def row_description(self, row: dict) -> str:
        try:
            return self.row_description_template.format(**row)
        except (KeyError, ValueError):
            return ""


class FormatterChoices(models.TextChoices):
    PLAINTEXT = "plaintext", _("Plaintext")
    FLOAT = "float", _("Float")
    INTEGER = "integer", _("Integer")
    DATE = "date", _("Date")
    DATETIME = "datetime", _("Date and time")
    BOOLEAN = "boolean", _("Boolean")
    LINK = "link", _("Link")
    SUMMARY = "summary", _("Summary")
    ABBREVIATION = "abbreviation", _("Abbreviation")
    IFRAME = "iframe", _("IFrame")


class FilterChoices(models.TextChoices):
    ICONTAINS = "ilike", _("Substring")
    INTEGER_RANGE = "integer_range", _("Integer Range")


class NonStrippingCharField(models.CharField):
    """A TextField that does not strip whitespace at the beginning/end of
    it's value. Might be important for markup/code."""

    def formfield(self, **kwargs):
        kwargs["strip"] = False
        return super().formfield(**kwargs)


class Column(models.Model):
    table = models.ForeignKey(
        Table,
        on_delete=models.CASCADE,
        related_name="columns",
        verbose_name=_("Table"),
    )
    name = models.CharField(
        max_length=255,
        verbose_name=_("Name"),
        help_text=_("Internal name of the column."),
    )
    label = models.CharField(
        max_length=255,
        verbose_name=_("Label"),
        help_text=_("Displayed label of the column."),
    )
    order = models.PositiveIntegerField(
        default=0,
        blank=False,
        null=False,
        verbose_name=_("Order"),
        help_text=_("Order of the column."),
    )
    visible = models.BooleanField(
        default=True,
        verbose_name=_("Visible"),
        help_text=_("Indicates whether the column is visible in table view."),
    )
    visible_detail = models.BooleanField(
        default=True,
        verbose_name=_("Visible in detail"),
        help_text=_("Indicates whether the column is visible in detail view."),
    )
    sortable = models.BooleanField(
        default=False,
        verbose_name=_("Sortable"),
        help_text=_("Indicates whether the column is sortable."),
    )
    searchable = models.BooleanField(
        default=False,
        verbose_name=_("Searchable"),
        help_text=_("Indicates whether the column is included in search queries."),
    )
    facet_count = models.PositiveSmallIntegerField(
        default=0,
        verbose_name=_("Facet Count"),
        help_text=_("Number of facet entries, 0 if disabled."),
    )
    prefix = NonStrippingCharField(
        max_length=255,
        blank=True,
        verbose_name=_("Prefix"),
        help_text=_("Prefix when outputting the column value."),
    )
    postfix = NonStrippingCharField(
        max_length=255,
        blank=True,
        verbose_name=_("Postfix"),
        help_text=_("Postfix when outputting the column value."),
    )
    formatter = models.CharField(
        choices=FormatterChoices.choices,
        max_length=255,
        default="plaintext",
        verbose_name=_("Formatter"),
        help_text=_("Formatter for the column."),
    )
    formatter_arguments = models.JSONField(
        blank=True,
        default=dict,
        verbose_name=_("Formatter Arguments"),
        help_text=_("Arguments for the formatter."),
    )
    filter = models.CharField(
        choices=FilterChoices.choices,
        max_length=255,
        default="",
        blank=True,
        verbose_name=_("Filter"),
        help_text=_("Filter function for the column."),
    )
    filter_arguments = models.JSONField(
        blank=True,
        default=dict,
        verbose_name=_("Filter Arguments"),
        help_text=_("Arguments for the selected filter."),
    )

    class Meta:
        verbose_name = _("Column")
        verbose_name_plural = _("Columns")
        ordering = ("order",)
        constraints = [
            models.UniqueConstraint(
                fields=["table", "name"], name="unique_table_column_name"
            ),
        ]

    def __str__(self):
        return f"{self.label} - {self.table}"



import csv

from django.http import Http404, StreamingHttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.views.generic import ListView

from .forms import FilterForm
from .models import Dataset, Table
from .table import RowQueryset, get_facets, get_row


def get_viewable_datasets(request, queryset):
    if not request.user.has_perm("datashow.view_dataset"):
        queryset = queryset.filter(public=True)
    return queryset


def get_dataset(request, slug: str) -> Dataset:
    queryset = get_viewable_datasets(request, Dataset.objects.all())
    return get_object_or_404(queryset.select_related("default_table"), slug=slug)


def get_table(dataset: Dataset, table_slug: str) -> Table:
    return get_object_or_404(
        dataset.tables.all().select_related("primary_key"), slug=table_slug
    )


def dataset_index(request, slug: str):
    dataset = get_dataset(request, slug)
    if dataset.default_table:
        return default_table_view(request, dataset=dataset, table=dataset.default_table)

    tables = dataset.tables.all().filter(visible=True)
    return render(
        request, "datashow/dataset_index.html", {"dataset": dataset, "tables": tables}
    )


class RowList(ListView):
    template_name = "datashow/dataset_table.html"
    context_object_name = "tables"
    dataset_slug_kwarg = "slug"
    table_slug_kwarg = "table_slug"
    default_view = False

    def get_queryset(self):
        self.dataset = self.kwargs.get("dataset")
        if self.dataset is None:
            dataset_slug = self.kwargs.get(self.dataset_slug_kwarg)
            self.dataset = get_dataset(self.request, dataset_slug)
        self.table = self.kwargs.get("table")
        if self.table is None:
            table_slug = self.kwargs.get(self.table_slug_kwarg)
            self.table = get_table(self.dataset, table_slug)

        self.filter_form = FilterForm(self.table, data=self.request.GET)
        self.formdata = None
        if self.filter_form.is_valid():
            self.formdata = self.filter_form.cleaned_data
        return RowQueryset(self.table, formdata=self.formdata)

    def render_to_response(self, context):
        if not self.default_view and self.dataset.default_table == self.table:
            return redirect(self.table.get_absolute_url())
        return super().render_to_response(context)

    def get_paginate_by(self, queryset):
        return self.table.pagination_size

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["dataset"] = self.dataset
        ctx["table"] = self.table
        ctx["columns"] = self.table.get_formatted_columns()
        ctx["facets"] = get_facets(self.table, self.formdata)
        ctx["filter_form"] = self.filter_form
        return ctx

    def get_template_names(self) -> list[str]:
        if self.request.headers.get("hx-boosted", "") == "true":
            return ["datashow/_table.html"]
        return super().get_template_names()


default_table_view = RowList.as_view(default_view=True)


class Echo:
    """An object that implements just the write method of the file-like
    interface.
    """

    def __init__(self, header_cols):
        self.header_cols = header_cols
        self.wrote_header = False

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        if not self.wrote_header:
            self.wrote_header = True
            return ",".join(self.header_cols) + "\n" + value
        return value


def table_csv_export(request, slug, table_slug):
    """A view that streams the filtered/sorted table as a CSV file."""

    dataset = get_dataset(request, slug)
    table = get_table(dataset, table_slug)

    filter_form = FilterForm(table, data=request.GET)
    formdata = None
    if filter_form.is_valid():
        formdata = filter_form.cleaned_data
    generator = RowQueryset(table, formdata=formdata).stream_raw()

    pseudo_buffer = Echo(table.get_sql_columns())
    writer = csv.writer(pseudo_buffer)

    return StreamingHttpResponse(
        (writer.writerow(row) for row in generator),
        content_type="text/csv",
        headers={"Content-Disposition": 'attachment; filename="%s.csv"' % table.slug},
    )


def show_dataset_default_table_row(request, slug: str, row_slug: str):
    dataset = get_dataset(request, slug)
    if not dataset.default_table:
        raise Http404
    return dataset_row(request, dataset, dataset.default_table, row_slug)


def show_dataset_table_row(request, slug: str, table_slug: str, row_slug: str):
    dataset = get_dataset(request, slug)
    table = get_table(dataset, table_slug)
    if dataset.default_table == table:
        return redirect("datashow:dataset-row", slug=dataset.slug, row_slug=row_slug)
    return dataset_row(request, dataset, table, row_slug)


def dataset_row(request, dataset: Dataset, table: Table, row_slug: str):
    if not table.primary_key:
        raise Http404
    try:
        row, row_dict = get_row(table, row_slug)
    except KeyError:
        raise Http404
    row_label = table.row_label(row_dict)
    row_description = table.row_description(row_dict)
    return render(
        request,
        "datashow/dataset_row.html",
        {
            "dataset": dataset,
            "table": table,
            "row_label": row_label,
            "row_description": row_description,
            "row": row,
            "object": row_dict,
        },
    )



import re
from typing import Optional

from .forms import SEARCH_PARAM, SORT_PARAM
from .models import FilterChoices

NONE_VALUE = "-"
SEP = re.compile(r'\s+|(".*?")')
MIN_WORD_PREFIX_LENGTH = 3
PREFIX_QUERY_MARKER = "*"


def make_fts_query(q: str) -> str:
    if q.count('"') % 2:
        q += '"'
    query = []
    for p in (x for x in SEP.split(q) if x and x != '""'):
        if '"' not in p and len(p) >= MIN_WORD_PREFIX_LENGTH:
            p = f'"{p}"*'
        elif '"' not in p:
            p = f'"{p}"'
        query.append(p)
    return " ".join(query)


class SqlQuery:
    def __init__(self, table, formdata: Optional[dict] = None):
        self.table = table
        self.column_map = {col.name: col for col in table.get_columns()}
        self.select = []
        self._from = [self.table.as_sql()]
        self.where = []
        self.groupby = []
        self.order = []
        self.limit = None
        self.offset = None
        self.params = []
        self.formdata = formdata

    def __str__(self) -> str:
        return "{} {}".format(self.to_sql(), self.params)

    def add_select(self):
        self.select = [
            '"{table}"."{col}" AS "{col}"'.format(table=self.table.name, col=col)
            for col in self.table.get_sql_columns()
        ]

    def add_limit(self, page_slice):
        if page_slice:
            self.limit = page_slice.stop - (page_slice.start or 0)
            self.offset = page_slice.start or 0
        else:
            self.limit = self.table.pagination_size

    def with_list(self, page_slice=None):
        self.add_select()
        self.add_filter()
        self.add_order()
        self.add_limit(page_slice)
        return self

    def with_all(self):
        self.add_select()
        self.add_filter()
        self.add_order()
        return self

    def with_count(self):
        self.select = ["COUNT(*)"]
        self.add_filter()
        return self

    def with_row(self, row_id):
        self.add_select()
        self.where.append(
            '"{table}"."{pk_col}" = ?'.format(
                table=self.table.name, pk_col=self.table.primary_key.name
            )
        )
        self.params.append(row_id)
        return self

    def with_facet(self, column):
        self.select = [
            '"{table}"."{col}" AS key, COUNT(*) AS count'.format(
                table=self.table.name, col=column.name
            )
        ]
        self.add_filter(exclude_column=column.name)
        self.limit = column.facet_count
        if column.sortable:
            self.order = ["key ASC"]
        else:
            self.order = ["count DESC"]
        self.groupby = ["key"]
        return self

    def to_sql(self) -> str:
        return "SELECT {select} FROM {from_} {where} {groupby} {order} {limit} {offset}".format(
            select=", ".join(self.select),
            # join JOINs with space
            from_=" ".join(self._from),
            where="WHERE " + " AND ".join(self.where) if self.where else "",
            groupby="GROUP BY " + ", ".join(self.groupby) if self.groupby else "",
            order="ORDER BY " + ", ".join(self.order) if self.order else "",
            limit="LIMIT {}".format(self.limit) if self.limit else "",
            offset="OFFSET {}".format(self.offset) if self.offset else "",
        )

    def add_order(self):
        if not self.formdata:
            return
        sort_columns = self.formdata.get(SORT_PARAM, [])
        for sort in sort_columns:
            direction = "ASC"
            if sort.startswith("-"):
                direction = "DESC"
                sort = sort[1:]
            if sort not in self.column_map:
                continue
            if not self.column_map[sort].sortable:
                continue
            self.order.append(
                '"{table}"."{col}" {dir}'.format(
                    table=self.table.name, col=sort, dir=direction
                )
            )

    def add_filter(self, exclude_column=None):
        if not self.formdata:
            return
        filter_columns = self.table.get_filter_columns()
        filter_columns = [col for col in filter_columns if col.name != exclude_column]
        facet_columns = self.table.get_facet_columns()
        facet_columns = [col for col in facet_columns if col.name != exclude_column]

        query: str = self.formdata.get(SEARCH_PARAM)
        if query and self.table.has_fts():
            query = make_fts_query(query)
            fts_table_name = f"{self.table.name}__fts"
            pk_col = self.table.primary_key.name
            self._from.append(
                f'JOIN "{fts_table_name}" ON "{self.table.name}"."{pk_col}" = "{fts_table_name}"."{pk_col}"'
            )
            self.where.append(f"{fts_table_name} MATCH ?")
            self.params.append(query)

        for column in facet_columns:
            value = self.formdata.get(column.name)
            if value:
                if value == NONE_VALUE:
                    self.where.append(f'"{self.table.name}"."{column.name}" IS NULL')
                else:
                    self.where.append(f'"{self.table.name}"."{column.name}" = ?')
                    self.params.append(value)

        for column in filter_columns:
            column_filter = column.filter
            if column_filter == FilterChoices.INTEGER_RANGE:
                field_value = self.formdata.get(column.name + "__range")
                if field_value is None:
                    continue
                min_value, max_value = field_value
                if min_value is not None and max_value is not None:
                    self.where.append(
                        f'"{self.table.name}"."{column.name}" BETWEEN ? AND ?'
                    )
                    self.params.extend([min_value, max_value])
                elif min_value is not None:
                    self.where.append(f'"{self.table.name}"."{column.name}" >= ?')
                    self.params.append(min_value)
                elif max_value is not None:
                    self.where.append(f'"{self.table.name}"."{column.name}" <= ?')
                    self.params.append(max_value)



import re
from datetime import datetime
from typing import Union

from django.contrib.humanize.templatetags.humanize import intcomma
from django.utils import formats
from django.utils.html import escape, format_html
from django.utils.safestring import SafeString, mark_safe
from django.utils.translation import gettext_lazy as _

from .models import FormatterChoices

TRAILING_ZERO = re.compile(r"[,\.]0+$")


def render_value(column, value):
    return format_html(
        "{prefix}{value}{postfix}",
        prefix=column.prefix,
        value=value,
        postfix=column.postfix,
    )


def try_format(args, key, row_data, default=""):
    try:
        value = args[key]
        return value.format(**row_data)
    except (KeyError, ValueError):
        return default


def generate_formatted_attrs(args, row_data):
    return mark_safe(
        " ".join(
            '{}="{}"'.format(escape(k), escape(try_format(args, k, row_data)))
            for k in args
        )
    )


ALIGN_RIGHT = "text-end"
ALIGN_CENTER = "text-center"
NUMBER = "tabular-numbers text-end"


def format_column(column):
    css = ""
    formatter = column.formatter
    if formatter == FormatterChoices.FLOAT:
        css = ALIGN_RIGHT
    elif formatter == FormatterChoices.INTEGER:
        css = ALIGN_RIGHT
    elif formatter == FormatterChoices.DATE:
        css = ALIGN_RIGHT
    elif formatter == FormatterChoices.DATETIME:
        css = ALIGN_RIGHT
    elif formatter == FormatterChoices.BOOLEAN:
        css = ALIGN_CENTER
    return css


FormattedValue = tuple[str, Union[str, SafeString]]


def format_link(value, args, row_data):
    attrs = generate_formatted_attrs(args, row_data)
    return format_html(
        "<a {attrs}>{link}</a>",
        attrs=attrs,
        link=value,
    )


def format_value(column, value, row_data, detail=False) -> FormattedValue:
    css = ""
    formatter = column.formatter
    args = column.formatter_arguments
    if value is None:
        value = mark_safe('<span class="text-secondary">–</span>')
        if formatter == FormatterChoices.BOOLEAN:
            css = ALIGN_CENTER
        if formatter in (
            FormatterChoices.FLOAT,
            FormatterChoices.INTEGER,
            FormatterChoices.DATE,
            FormatterChoices.DATETIME,
        ):
            css = ALIGN_RIGHT
        if formatter == FormatterChoices.LINK:
            return css, render_value(column, format_link(value, args, row_data))
        return css, value

    if formatter == FormatterChoices.FLOAT:
        value = intcomma(value)
        css = NUMBER
    elif formatter == FormatterChoices.INTEGER:
        value = TRAILING_ZERO.sub("", intcomma(value))
        css = NUMBER
    elif formatter == FormatterChoices.DATE:
        value = formats.date_format(datetime.fromisoformat(value), "SHORT_DATE_FORMAT")
        css = NUMBER
    elif formatter == FormatterChoices.DATETIME:
        value = formats.date_format(
            datetime.fromisoformat(value), "SHORT_DATETIME_FORMAT"
        )
        css = NUMBER
    elif formatter == FormatterChoices.BOOLEAN:
        if value:
            value = format_html(
                '<span class="text-success">✅ {}</span>',
                _("Yes"),
            )
        else:
            value = format_html(
                '<span class="text-danger">❌ {}</span>',
                _("No"),
            )
        css = ALIGN_CENTER
    elif formatter == FormatterChoices.LINK:
        value = format_link(value, args, row_data)
    elif formatter == FormatterChoices.SUMMARY:
        if not detail:
            summary = try_format(args, "summary", row_data, _("Details"))
            value = format_html(
                '<details class="datashow-summary"><summary><span>{summary}</span></summary>{details}</details>',
                summary=summary,
                details=value,
            )
    elif formatter == FormatterChoices.ABBREVIATION:
        title = try_format(args, "title", row_data, None)
        if title is None:
            value = format_html("<abbr>{value}</abbr>", value=value)
        else:
            value = format_html(
                '<abbr title="{title}">{value}</abbr>',
                title=title,
                value=value,
            )
    elif formatter == FormatterChoices.IFRAME:
        attrs = generate_formatted_attrs(args, row_data)
        value = format_html(
            '<iframe src="{url}" {attrs}></iframe>', url=value, attrs=attrs
        )

    return css, render_value(column, value)
